#include <iostream>
#include <climits>
#include <fstream>
#include "libastar.h"

LibAstar::LibAstar(nodeMatrix* map) : start(map->at(0).at(0)), end(map->at(map->size() - 1).at(
        map->at(map->size() - 1).size() - 1)), map(map) {}

nodeList LibAstar::findPath() {
    // Find path using A* algorithm
    Node* currentNode = start;
    currentNode->setCostG(0);
    currentNode->setCostH(distance(currentNode, end));
    currentNode->setCostF(currentNode->getCostG() + currentNode->getCostH());
    currentNode->setParent(nullptr);
    open_list.push_back(currentNode);
    closed_list.push_back(currentNode);
    while (currentNode->getX() != end->getX() || currentNode->getY() != end->getY()) {
        // Find children of current node
        nodeList children = findChildren(currentNode);
        for (Node* child: children) {
            if (child != nullptr) {
                // Add child to open list if it is not already on it
                if (!isNodeOnList(child, closed_list)) {
                    child->setCostG(currentNode->getCostG() + distance(currentNode, child));
                    child->setCostH(distance(child, end));
                    child->setCostF(child->getCostG() + child->getCostH());
                    child->setParent(currentNode);
                    if (isNodeOnList(child, open_list)) {
                        int place = findPlace(child, open_list);
                        if (child->getCostF() < open_list[place]->getCostF()) {
                            open_list[place] = child;
                        }
                    } else {
                        open_list.push_back(child);
                    }
                }
            }
        }
        // Find node with the lowest cost
        currentNode = findLowestCostNode(open_list);
        if (currentNode == nullptr) {
            return {};
        }
        // Add current node to closed list
        closed_list.push_back(currentNode);
        // Remove current node from open list
        int place = findPlace(currentNode, open_list);
        open_list.erase(open_list.begin() + place);
    }
    // Path found
    // reconstruct path
    currentNode = closed_list[closed_list.size() - 1];
    while (currentNode != nullptr) {
        path.push_back(currentNode);
        currentNode = currentNode->getParent();
    }
    return path;
}


nodeList LibAstar::findChildren(Node* pNode) {
    nodeList children;
    int x = pNode->getX();
    int y = pNode->getY();
    if (x > 0) {
        children.push_back(map->at(x - 1).at(y));
    }
    if (x < map->size() - 1) {
        children.push_back(map->at(x + 1).at(y));
    }
    if (y > 0) {
        children.push_back(map->at(x).at(y - 1));
    }
    if (y < map->at(x).size() - 1) {
        children.push_back(map->at(x).at(y + 1));
    }
    if (x > 0 && y > 0) {
        children.push_back(map->at(x - 1).at(y - 1));
    }
    if (x > 0 && y < map->at(x).size() - 1) {
        children.push_back(map->at(x - 1).at(y + 1));
    }
    if (x < map->size() - 1 && y > 0) {
        children.push_back(map->at(x + 1).at(y - 1));
    }
    if (x < map->size() - 1 && y < map->at(x).size() - 1) {
        children.push_back(map->at(x + 1).at(y + 1));
    }
    for (int i = 0; i < children.size(); i++) {
        if (children[i]->getObstacle() == 1) {
            children.erase(children.begin() + i);
            i--;
        }
    }
    return children;
}

bool LibAstar::isNodeOnList(Node* pNode, const nodeList &vector1) {
    for (Node* node: vector1) {
        if (node->getX() == pNode->getX() && node->getY() == pNode->getY()) {
            return true;
        }
    }
    return false;
}

int LibAstar::findPlace(Node* pNode, nodeList vector) {
    for (int i = 0; i < vector.size(); i++) {
        if (vector[i] == pNode) {
            return i;
        }
    }
    return -1;
}

Node* LibAstar::findLowestCostNode(const nodeList &vector) {
    Node* lowestCostNode = nullptr;
    int lowestCost = INT_MAX;
    for (Node* node: vector) {
        if (node->getCostF() < lowestCost) {
            lowestCost = node->getCostF();
            lowestCostNode = node;
        }
    }
    return lowestCostNode;
}

int verifyPath(nodeMatrix* map, nodeList* path, nodeList* astarPath) {
    // Check start
    if (path->at(path->size() - 1)->getX() != 0 || path->at(path->size() - 1)->getY() != 0) {
        std::cout << "\033[31mStart is not at the beginning of the map\033[0m" << std::endl;
        return -1;
    }
    // Check end
    if (path->at(0)->getX() != map->size() - 1 || path->at(0)->getY() != map->at(0).size() - 1) {
        std::cout << "\033[31mEnd is not at the end of the map\033[0m" << std::endl;
        return -1;
    }
    // Check path
    for (int i = 0; i < path->size() - 1; i++) {
        // Check if path is blocked
        if (path->at(i)->getX() == path->at(i + 1)->getX() && path->at(i)->getY() == path->at(i + 1)->getY()) {
            std::cout << "\033[31mPath is blocked\033[0m" << std::endl;
            return -1;
        }
        // Check if path is continuous
        if (distance(path->at(i), path->at(i + 1)) > 2) {
            std::cout << "\033[31mPath is not continuous\033[0m" << std::endl;
            return -1;
        }
        // Check if node is an obstacle
        if (map->at(path->at(i)->getX()).at(path->at(i)->getY())->getObstacle() == 1) {
            std::cout << "\033[31mNode is an obstacle\033[0m" << std::endl;
            return -1;
        }
    }
    float relativeError = abs(path->size() - astarPath->size()) / astarPath->size();
    if (relativeError > 0.1) {
        std::cout << "\033[31mPath is not optimal\033[0m" << std::endl;
        return -1;
    }
    return 0;
}

/**
 * Print map and path in ppm format
 * @param map
 * @param path
 */
void printPath(nodeMatrix* map, nodeList* path, nodeList* openList, nodeList* closedList) {
    std::ofstream file;
    file.open("map.ppm");
    file << "P3\n" << map->size() << " " << map->at(0).size() << "\n255\n";
    for (auto & i : *map) {
        for (auto & j : i) {
            if (LibAstar::isNodeOnList(j, *path)) {
                file << "255 0 0 ";
            } else if (LibAstar::isNodeOnList(j, *openList)) {
                file << "0 0 255 ";
            } else if (LibAstar::isNodeOnList(j, *closedList)) {
                file << "0 255 0 ";
            } else if (j->getObstacle() == 1) {
                file << "0 0 0\n";
            } else {
                file << "255 255 255\n";
            }
        }
    }
    file.close();
}
