#ifndef ASTAR_WORLD_HPP
#define ASTAR_WORLD_HPP
#include <vector>
#include <string>

/**
 * @class Node
 * @brief A node in the world.
 * This class represents a node in the world with a position, a cost, a parent and a state.
 */
class Node{
private:
    int x;
    int y;
    int cost_g;
    int cost_h;
    int cost_f;
    int obstacle;
    Node *parent;
public:
    Node(int x, int y, int obstacle);
    int getX() const;
    int getY() const;
    int getCostG() const;
    int getCostH() const;
    int getCostF() const;
    int getObstacle() const;
    Node *getParent();
    void setCostG(int costG);
    void setCostH(int costH);
    void setCostF(int costF);
    void setParent(Node *newParent);
    ~Node()=default;
};

/**
 * @brief Computes the distance between two nodes.
 * @param node1 The first node.
 * @param node2 The second node.
 * @return The distance between the two nodes (as integer).
 */
int distance(Node *node1, Node *node2);

/**
 * Creates a world with a given size.
 * @param width the width of the world.
 * @param height the height of the world.
 * @param obstacleProbability the probability of an obstacle in a cell.
 * @return a vector of vectors of nodes.
 */
std::vector<std::vector<Node*>> createMap(int width, int height, int obstacleProbability, bool hard);

/**
 * Calculate the density of the obstacles in the world.
 * @param map a pointer to the map.
 * @return a float representing the density of the obstacles.
 */
float density(std::vector<std::vector<Node*>>* map);

/**
 * Free the memory of the map.
 * @param map a pointer to the map.
 */
void deleteMap(std::vector<std::vector<Node*>>* map);
#endif //ASTAR_WORLD_HPP
